"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/* eslint-disable functional/prefer-immutable-types */
const type_utils_1 = require("@typescript-eslint/type-utils");
const is_immutable_type_1 = require("is-immutable-type");
const typescript_1 = require("typescript");
const common_1 = require("./common");
const unsafe_assignment_rule_1 = require("./unsafe-assignment-rule");
/**
 * An ESLint rule to ban unsafe assignment from mutable to readonly types.
 */
// eslint-disable-next-line total-functions/no-unsafe-readonly-mutable-assignment
const noUnsafeMutableReadonlyAssignment = (0, common_1.createRule)({
    name: "no-unsafe-mutable-readonly-assignment",
    meta: {
        type: "problem",
        docs: {
            description: "Bans unsafe assignment from mutable to readonly types.",
            recommended: "error",
        },
        messages: {
            errorStringGeneric: 'Unsafe "{{ sourceImmutability }}" to "{{ destinationImmutability }}" assignment. Source: "{{ sourceType }}", destination: "{{ destinationType }}"',
        },
        schema: [],
    },
    create: (0, unsafe_assignment_rule_1.createNoUnsafeAssignmentRule)((checker, rawDestinationType, rawSourceType, sourceNode) => {
        // eslint-disable-next-line functional/no-conditional-statements
        if ((0, common_1.isLiteral)(sourceNode)) {
            return "safe";
        }
        const typePairs = (0, common_1.assignableTypePairs)(rawDestinationType, rawSourceType, checker);
        // TODO support config
        const overrides = (0, is_immutable_type_1.getDefaultOverrides)();
        const allSafe = typePairs.every(({ sourceType, destinationType }) => {
            // eslint-disable-next-line functional/no-conditional-statements
            if (sourceType === destinationType) {
                // safe
                return true;
            }
            // eslint-disable-next-line functional/no-conditional-statements
            if ((0, type_utils_1.isTypeAnyType)(sourceType) || (0, type_utils_1.isTypeAnyType)(destinationType)) {
                // safe
                return true;
            }
            // eslint-disable-next-line functional/no-conditional-statements
            if ((0, type_utils_1.isTypeUnknownType)(sourceType) ||
                (0, type_utils_1.isTypeUnknownType)(destinationType)) {
                // safe
                return true;
            }
            // eslint-disable-next-line functional/no-conditional-statements
            if ((0, type_utils_1.isTypeFlagSet)(destinationType, typescript_1.TypeFlags.VoidLike)) {
                // safe
                return true;
            }
            const destinationImmutability = (0, is_immutable_type_1.getTypeImmutability)(checker, destinationType, overrides);
            const sourceImmutability = (0, is_immutable_type_1.getTypeImmutability)(checker, sourceType, overrides);
            const isUnsafe = (destinationImmutability === is_immutable_type_1.Immutability.Immutable ||
                destinationImmutability === is_immutable_type_1.Immutability.ReadonlyDeep) &&
                (sourceImmutability === is_immutable_type_1.Immutability.ReadonlyShallow ||
                    sourceImmutability === is_immutable_type_1.Immutability.Mutable);
            return !isUnsafe;
        });
        return allSafe ? "safe" : "unsafe";
    }),
    defaultOptions: [],
});
exports.default = noUnsafeMutableReadonlyAssignment;
